// cl_LF internals

#ifndef _CL_LF_H
#define _CL_LF_H

#include "cl_number.h"

struct cl_heap_lfloat : cl_heap {
	unsigned int len;	// length of mantissa (in digits)
	int sign;		// sign (0 or -1)
	uint32 expo;		// exponent
	uintD data[1];		// mantissa
};

// Minimum number of mantissa digits,
// so that a LF has not fewer mantissa bits than a DF.
  #define LF_minlen  ceiling(53,intDsize)
// Exponent.
  #define LF_exp_low  1
  #define LF_exp_mid  0x80000000UL
  #define LF_exp_high 0xFFFFFFFFUL

inline cl_heap_lfloat* TheLfloat (cl_heap_lfloat* p)
	{ return p; }
inline cl_heap_lfloat* TheLfloat (const cl_number& obj)
	{ return (cl_heap_lfloat*)(obj.pointer); }


// Liefert zu einem Long-Float x : (futruncate x), ein LF.
// x wird von der 0 weg zur nchsten ganzen Zahl gerundet.
extern cl_LF futruncate (const cl_LF& x);

// shorten(x,len) verkrzt ein Long-Float x auf gegebene Lnge len
// und rundet dabei.
// > cl_LF x: ein Long-FLoat
// > uintC len: gewnschte Lnge (>= LF_minlen, < TheLfloat(x)->len)
// < cl_LF ergebnis: verkrztes Long-Float
extern cl_LF shorten (const cl_LF& x, uintC len);

// extend(x,len) verlngert ein Long-Float x auf gegebene Lnge len.
// > cl_LF x: ein Long-FLoat
// > uintC len: gewnschte Lnge (> TheLfloat(x)->len)
// < cl_LF ergebnis: verlngertes Long-Float
extern cl_LF extend (const cl_LF& x, uintC len);

// LF_to_LF(x,len) wandelt ein Long-Float x in ein Long-Float gegebener Lnge
// len um und rundet dabei ntigenfalls.
// > cl_LF x: ein Long-FLoat
// > uintC len: gewnschte Lnge (>= LF_minlen)
// < cl_LF ergebnis: Long-Float gegebener Lnge
extern cl_LF LF_to_LF (const cl_LF& x, uintC len);

// GEN_LF_OP2(arg1,arg2,LF_OP,ergebnis_zuweisung)
// generates the body of a LF operation with two arguments.
// LF_OP is only executed once both arguments have been converted to the same
// float format (the longer one of arg1 and arg2). The result is then
// converted the shorter of the two float formats.
#define GEN_LF_OP2(arg1,arg2,LF_OP,ergebnis_zuweisung)  \
{									\
	var uintC len1 = TheLfloat(arg1)->len;				\
	var uintC len2 = TheLfloat(arg2)->len;				\
	if (len1==len2) /* gleich -> direkt ausfhren */		\
		return LF_OP(arg1,arg2);				\
	elif (len1>len2) /* -> arg2 auf die Lnge von arg1 bringen */	\
		return shorten(LF_OP(arg1,extend(arg2,len1)),len2);	\
	else /* (len1<len2) -> arg1 auf die Lnge von arg2 bringen */	\
		return shorten(LF_OP(extend(arg1,len2),arg2),len1);	\
}

// Liefert zu zwei gleichlangen Long-Float x und y : (+ x y), ein LF.
// LF_LF_plus_LF(x)
extern cl_LF LF_LF_plus_LF (const cl_LF& x, const cl_LF& y);

// Liefert zu zwei gleichlangen Long-Float x und y : (- x y), ein LF.
// LF_LF_minus_LF(x)
extern cl_LF LF_LF_minus_LF (const cl_LF& x, const cl_LF& y);

// Use this macro if ALL of your cl_LF operations (+, -, *, /) in the
// rest of your file ALWAYS get two operands of the same precision.
#define ALL_cl_LF_OPERATIONS_SAME_PRECISION()  \
									\
inline cl_LF operator+ (const cl_LF& x, const cl_LF& y)			\
{									\
	return LF_LF_plus_LF(x,y);					\
}									\
									\
inline cl_LF operator- (const cl_LF& x, const cl_LF& y)			\
{									\
	return LF_LF_minus_LF(x,y);					\
}

// LF_to_I(x) wandelt ein Long-Float x, das eine ganze Zahl darstellt,
// in ein Integer um.
extern cl_I cl_LF_to_I (const cl_LF& x);

// cl_I_to_LF(x,len) wandelt ein Integer x in ein Long-Float um und rundet dabei.
extern cl_LF cl_I_to_LF (const cl_I& x, uintC len);

// cl_RA_to_LF(x,len) wandelt eine rationale Zahl x in ein Long-Float um
// und rundet dabei.
extern cl_LF cl_RA_to_LF (const cl_RA& x, uintC len);

// LF_I_mul(x,y) multipliziert ein Long-Float x und ein Integer y.
extern cl_R cl_LF_I_mul (const cl_LF& x, const cl_I& y);

// LF_I_div(x,y) dividiert ein Long-Float x durch ein Integer y.
extern cl_LF cl_LF_I_div (const cl_LF& x, const cl_I& y);

// Vergrert eine Long-Float-Lnge n, so da aus d = intDsize*n
// mindestens d+sqrt(d)+2 wird.
extern uintC cl_LF_len_incsqrt (uintC len);

// Vergrert eine Long-Float-Lnge n, so da aus d = intDsize*n
// mindestens d+sqrt(d)+2+(LF_exp_len-1) wird.
extern uintC cl_LF_len_incsqrtx (uintC len);

// Subroutine for evaluating
// sum(0 <= n < N, a(n)/b(n) * (p(0)...p(n))/(q(0)...q(n)))
// where all the entries are small integers (ideally polynomials in n).
// This is fast because it groups factors together before multiplying.
// Arguments:
//   Vectors p[0..N-1], q[0..N-1], a[0..N-1], b[0..N-1], N.
//   Some of the vectors (a,b,q) can be a NULL pointer, all of its entries are
//   then understood to be 1.
//   If given, a vector qs[0..N-1] which the evaluation routine may use to
//   split off q[n] into q[n]*2^qs[n]. qs may be NULL, in that case no shift
//   optimizations will be used. (They are worth it only if a significant
//   amount of multiplication work can be saved by shifts.)
// Result will be a cl_LF with len digits.
struct cl_rational_series {
	// To be set explicitly.
	const cl_I* pv;
	      cl_I* qv;
	const cl_I* av;
	const cl_I* bv;
	     uintL* qsv;
};
extern cl_LF eval_rational_series (uintL N, const cl_rational_series& args, uintC len);

// In this alternate implementation the series is not represented as a couple
// of arrays, but as a method returning each tuple (p(n),q(n),a(n),b(n))
// in turn. This is preferrable if the a(n) are big, in order to avoid too
// much memory usage at the same time.
// Some of the factors (a,b) may be omitted. They are then understood to be 1.
// The next function is called N times and is expected to return
// (p(n),q(n),a(n),b(n)) for n=0..N-1 in that order.
//
struct cl_rational_series_pqab {
	cl_I p;
	cl_I q;
	cl_I a;
	cl_I b;
};
struct cl_rational_series_pqab_stream {
	cl_rational_series_pqab (*nextfn)(cl_rational_series_pqab_stream&);
	cl_rational_series_pqab next () { return nextfn(*this); }
	// Constructor.
	cl_rational_series_pqab_stream (cl_rational_series_pqab (*n)(cl_rational_series_pqab_stream&)) : nextfn (n) {}
};
extern cl_LF eval_rational_series (uintL N, cl_rational_series_pqab_stream& args, uintC len);
//
struct cl_rational_series_pqb {
	cl_I p;
	cl_I q;
	cl_I b;
};
struct cl_rational_series_pqb_stream {
	cl_rational_series_pqb (*nextfn)(cl_rational_series_pqb_stream&);
	cl_rational_series_pqb next () { return nextfn(*this); }
	// Constructor.
	cl_rational_series_pqb_stream (cl_rational_series_pqb (*n)(cl_rational_series_pqb_stream&)) : nextfn (n) {}
};
extern cl_LF eval_rational_series (uintL N, cl_rational_series_pqb_stream& args, uintC len);
//
struct cl_rational_series_pqa {
	cl_I p;
	cl_I q;
	cl_I a;
};
struct cl_rational_series_pqa_stream {
	cl_rational_series_pqa (*nextfn)(cl_rational_series_pqa_stream&);
	cl_rational_series_pqa next () { return nextfn(*this); }
	// Constructor.
	cl_rational_series_pqa_stream (cl_rational_series_pqa (*n)(cl_rational_series_pqa_stream&)) : nextfn (n) {}
};
extern cl_LF eval_rational_series (uintL N, cl_rational_series_pqa_stream& args, uintC len);
//
struct cl_rational_series_pq {
	cl_I p;
	cl_I q;
};
struct cl_rational_series_pq_stream {
	cl_rational_series_pq (*nextfn)(cl_rational_series_pq_stream&);
	cl_rational_series_pq next () { return nextfn(*this); }
	// Constructor.
	cl_rational_series_pq_stream (cl_rational_series_pq (*n)(cl_rational_series_pq_stream&)) : nextfn (n) {}
};
extern cl_LF eval_rational_series (uintL N, cl_rational_series_pq_stream& args, uintC len);


#endif /* _CL_LF_H */
